(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uuniverse.ml,v 1.7 2014/02/09 17:27:28 fclement Exp $*)

let u_qual = Upervasives.qualify "Uuniverse";;

let u_invalid_arg = Upervasives.invalid_arg_qual u_qual;;

type 'a index = 'a Uindex.t;;

type 'a indices = ('a, 'a index) Hashtbl.t;;
type 'a elements = 'a array;;

type 'a t =
  {
   indices : 'a indices;
   mutable elements : 'a elements;
   mutable cardinal : Unatural.t;
  }
;;

(* Expansion phase, or transient state. *)

let make n x =
  if n = Unatural._0_ then u_invalid_arg "make" "empty universe" else
  let indices = Hashtbl.create (Unatural.to_int n) in
  Hashtbl.add indices x (Uindex.of_natural Unatural._0_);
  let elements = Uarray.make n x in
  let cardinal = Unatural._1_ in
  { indices = indices; elements = elements; cardinal = cardinal; }
;;

let mem u x = Hashtbl.mem u.indices x;;

let index u x = Hashtbl.find u.indices x;;

let unsafe_add u x =
  let len = Uarray.length u.elements in
  let index = Uindex.of_natural u.cardinal in
  if Uindex.to_natural index = len then
  begin
    let elements =
      Uarray.make
        (Unatural.succ (Unatural.mul Unatural._2_ len)) x in
    Array.blit u.elements 0 elements 0 (Unatural.to_int u.cardinal);
    u.elements <- elements;
  end;
  Hashtbl.add u.indices x index;
  u.elements.(Uindex.to_int index) <- x;
  u.cardinal <- Unatural.succ u.cardinal;
;;

let add u x = if not (mem u x) then unsafe_add u x;;

let copy u =
  { u with
    indices = Hashtbl.copy u.indices;
    elements = Array.copy u.elements; }
;;

let product u1 u2 =
  let u10 = u1.elements.(0)
  and u20 = u2.elements.(0) in
  let u = make (Unatural.mul u1.cardinal u2.cardinal) (u10, u20) in
  let len1 = Unatural.pred (Uarray.length u1.elements) in
  let len2 = Unatural.pred (Uarray.length u2.elements) in
  let f x1 x2 = unsafe_add u (x1, x2) in
  Uarray.sub_iter (f u10) u2.elements 1 len2;
  Uarray.sub_iter (fun x1 -> Array.iter (f x1) u2.elements) u1.elements 1 len1;
  u
;;

let of_list l =
  let u = make (Ulist1.length l) (Ulist1.hd l) in
  Ulist1.iter (add u) l;
  u
;;

let of_array a =
  let len = Uarray.length a in
  if len = Unatural._0_ then u_invalid_arg "of_array" "empty array" else
  let u = make len a.(0) in
  Uarray.sub_iter (add u) a 1 (Unatural.pred len);
  u
;;

let gen_range succ i j =
  let rec loop accu i =
    if i >= j then Ulist.rev accu else
    loop (i :: accu) (succ i) in
  loop [] i
;;

let gen_of_range succ i j =
  if i >= j then u_invalid_arg "of_range" "empty range" else
  of_list (i, (gen_range succ) (succ i) j)
;;

let of_range = gen_of_range succ;;

let of_natural_range = gen_of_range Unatural.succ;;

let of_rectangle h w =
  let hu = of_range 0 h
  and wu = of_range 0 w in
  product hu wu
;;

let of_natural_rectangle h w =
  let hu = of_natural_range Unatural._0_ h
  and wu = of_natural_range Unatural._0_ w in
  product hu wu
;;

let indices u =
  gen_of_range Uindex.succ Uindex._0_ (Uindex.of_natural u.cardinal)
;;

(* Steady state. *)

let stats u = Hashtbl.stats u.indices;;

let cardinal u = u.cardinal;;

let valid_index (u : 'a t) (index : 'a index) =
  Uindex.to_natural index < cardinal u
;;

let check_index mod_name fun_name (u : 'a t) (index : 'a index) =
  if not (valid_index u index) then
    Upervasives.invalid_arg (Upervasives.qualify mod_name fun_name)
      (Printf.sprintf "index %d out of universe of %d element(s)"
         (Uindex.to_int index)
         (Unatural.to_int (cardinal u)))
;;

let element u index = u.elements.(Uindex.to_int index);;

let elements u = Array.sub u.elements 0 (Unatural.to_int u.cardinal);;

let sub_elements u sub_indices =
  Array.map (fun index -> element u index) sub_indices
;;

let fold_left f accu u = Uarray.sub_fold_left f accu u.elements 0 u.cardinal;;

let fold_left_i f accu u =
  Uarray.sub_fold_left_i
    (fun accu i -> f accu (Uindex.of_int i)) accu u.elements 0 u.cardinal
;;

let iter f u = Uarray.sub_iter f u.elements 0 u.cardinal;;

let iter_i f u =
  Uarray.sub_iter_i (fun i x -> f (Uindex.of_int i) x) u.elements 0 u.cardinal
;;

let map_elements f u =
  let res = Uarray.make u.cardinal (f u.elements.(0)) in
  Uarray.sub_iter_i
    (fun i x -> res.(i) <- f x)
    u.elements 1 (Unatural.pred u.cardinal);
  res
;;

let map f u = of_array (map_elements f u);;

let find p u = Uarray.sub_find p u.elements 0 u.cardinal;;

let find_all p u = Uarray.sub_find_all p u.elements 0 u.cardinal;;

let to_list u = Array.to_list (elements u);;
